import Lean
import Std

-- Structural Recursion
def replicate (n : Nat) (x : α) : List α :=
  match n with
  | 0 => []
  | n' + 1 => x :: replicate n' x

def reverse (xs : List α) : List α :=
  let rec rev (acc : List α) : List α → List α
    | [] => acc
    | y :: ys => rev (y :: acc) ys
  rev [] xs


-- Even Ackermann is primitive recursive at higher type!
def rep (f : Nat → Nat) (n : Nat) : Nat :=
  match n with
  | 0 => f 1
  | n' + 1 => f (rep f n')

def ack (n : Nat) : Nat → Nat :=
  match n with
  | 0 => (· + 1)
  | n' + 1 => rep (ack n')


-- Using a measure

def iota (n : Nat) : List Nat :=
  let rec iotaFrom (i : Nat) :=
    if i < n then
      i :: iotaFrom (i + 1)
    else []
  iotaFrom 0
termination_by
  iotaFrom i => n - i

namespace Demo

def Array.map (f : α → β) (arr : Array α) : Array β :=
  let rec aux (acc : Array β) (i : Nat) : Array β :=
    if h : i < arr.size then
      aux (acc.push (f arr[i])) (i + 1)
    else
      acc
  aux #[] 0
termination_by
  aux _ i => arr.size - i


def merge [Ord α] : List α → List α → List α
  | [], ys => ys
  | xs, [] => xs
  | x'::xs', y'::ys' =>
    match Ord.compare x' y' with
    | .lt | .eq => x' :: merge xs' (y' :: ys')
    | .gt => y' :: merge (x'::xs') ys'
termination_by merge xs ys => xs.length + ys.length

def splitList (lst : List α) : (List α × List α) :=
  match lst with
  | [] => ([], [])
  | x :: xs =>
    let (a, b) := splitList xs
    (x :: b, a)

theorem splitList_shorter_fst (lst : List α) (h : lst.length ≥ 2) :
    (splitList lst).fst.length < lst.length := by
  sorry

theorem splitList_shorter_snd (lst : List α) (h : lst.length ≥ 2) :
    (splitList lst).snd.length < lst.length := by
  sorry

def mergeSort [Ord α] (xs : List α) : List α :=
  if h : xs.length < 2 then
    match xs with
    | [] => []
    | [x] => [x]
  else
    let halves := splitList xs
    have : xs.length ≥ 2 := by
      apply Nat.ge_of_not_lt
      assumption
    have : halves.fst.length < xs.length := by
      apply splitList_shorter_fst
      assumption
    have : halves.snd.length < xs.length := by
      apply splitList_shorter_snd
      assumption
    merge (mergeSort halves.fst) (mergeSort halves.snd)
termination_by mergeSort xs => xs.length

#eval Lean.Meta.getEqnsFor? ``merge -- generates the lemmas


#check merge._eq_1
#check merge._eq_2
#check merge._eq_3


end Demo

inductive ParenExpr where
  | atom : String → ParenExpr
  | list : List ParenExpr → ParenExpr

def ParenExpr.toString (e : ParenExpr) : String :=
  match e with
  | .atom x => x
  | .list xs =>
    let strs := xs.attach.map (fun ⟨x, _⟩ => ParenExpr.toString x)
    "(" ++ String.join (strs.intersperse " ") ++ ")"
